/*-
 * Copyright (C)2008..2018 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2008..2018 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: style_pgo0.c,v 1.75 2018/12/13 15:55:43 babolo Exp $"

/*     NFS    HVUDLR */

#define BLIN_COMPAT      3
#define Bpars_COMPAT     3
#define MULAR_COMPAT     0
#define MIFE_COMPAT      4
#define PGOBLIN_COMPAT   2
#define PGOBLIN_INTERNAL 1
#define PGOBLIN_MODULE   pgoblin_module

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sysexits.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <multilar.h>
#include <mife.h>
#include "pgoblin.h"

#define PGOBLIN_NTBSIZE  8                      /*       0-255 */
#define PGOBLIN_STYLE_ID "#pGoblin-pgo0"

#define H 0
#define U 1
#define D 2
#define V 3
#define L 4
#define R 5
#define N 6 /*   NULL,    L  R                        */
            /*        HVUDLR:                  */
#define F 7 /*   NULL,    L  R                         */
#define S 7 /*    ,    L  R (?) */
#define M 7
#define ENTITY_LETTERS 27

typedef struct pgo0_arb {             /*                          */
    u_char  *nm;
    ssize_t  vl;
} pgo0_arb;

typedef union pgo0_term {             /*   (!arbi)                              */
    u_int32_t index;                  /*                                                  */
    struct {
        u_int32_t numb   : 8          /*  entity                                                */
                , dflt   : 1          /*                                                    */
                , dummy1 : 7
                , type   : 8          /*  entity (H, U, D, V, L, R, N = 0 .. 6)                   */
                , dummy0 : 6
                , sign   : 1          /*  (+, -)                                                 */
                , arbi   : 1          /*   !!!MOST SIGNIFICANT BIT!!!                */
#   define D_ARBI 0x80000000          /*                                                         */
        ;
    };
} pgo0_term;

typedef struct pgo0_ent {
    u_int32_t  def;                   /* 0  NULL     mular pgo0_style.str      */
    struct {
        u_int32_t  f;                 /* 0  NULL     mular pgo0_style.str      */
        u_int32_t  b;                 /* 0  NULL     mular pgo0_style.str      */
    } spec[1 << PGOBLIN_NTBSIZE];
} pgo0_ent;

typedef struct pgo0_style {
    char              id[PGOBLIN_STRING_ID_LEN];  /*  "#pGoblin"             */
    u_int32_t         flags;
#   define            PGOBLIN_NTBSZMK (~(~0U << PGOBLIN_NTBSIZE)) /* 0x0000FF */
#   define            FIRST                          0x000001
#   define            FORWARD (FIRST   << 1)      /* 0x000002 */
#   define            LAST                           0x000004
#   define            BAKWARD (LAST    << 1)      /* 0x000008 */
#   define            HSHIFT          12
#   define            VSHIFT          16
#   define            HFIRST  (FIRST   << HSHIFT) /* 0x001000   Output +H0 when zero rows            */
#   define            HHFIRST (FORWARD << HSHIFT) /* 0x002000 */
#   define            HLAST   (LAST    << HSHIFT) /* 0x004000   Output -H0 when zero rows            */
#   define            HHLAST  (BAKWARD << HSHIFT) /* 0x008000 */
#   define            VFIRST  (FIRST   << VSHIFT) /* 0x010000   Output +V0 when zero columns         */
#   define            VVFIRST (FORWARD << VSHIFT) /* 0x020000 */
#   define            VLAST   (LAST    << VSHIFT) /* 0x040000   Output -V0 when zero columns         */
#   define            VVLAST  (BAKWARD << VSHIFT) /* 0x080000 */
#   define            ALWAYS                         0x100000
#   define            HALWAYS  ALWAYS             /* 0x100000 */
#   define            VALWAYS (ALWAYS  << 1)      /* 0x200000 */
#   define            EDGE    (HFIRST | HLAST | VFIRST | VLAST)
/*
 * pgo0_style.flags
 * FIRST FORWARD
 *   0      0     @+0
 *   1      0   @+0   +0
 *   1      1   @+0  
 *
 * LAST  BAKWARD
 *   0      0     @-0
 *   1      0   @-0   -0
 *   1      1   @-0  
 */
#   define P_A (FIRST | LAST)
    mular_descriptor *str;  /*  -                             */
    mular_descriptor *arb;  /*                                                */
    mular_descriptor *rid;  /*                                                     */
    pgo0_ent          a[M];
} pgo0_style;

static const u_char entity[ENTITY_LETTERS + 1] =
{ M, M, M, M,  D, M, F, M,  H, M, M, M,  L, M, N, M
, M, M, R, S,  M, U, V, M,  M, M, M, M
};

static const u_int32_t flagvalues[M] = {HALWAYS, HFIRST, HLAST, VALWAYS, VFIRST, VLAST};
static const char *clasnames[] = {"L0", "Ll", "Lb", "Ln", "Lz", "Ls", "Lu", "Le", "Ld", "Lo", "Lx"};

#define L0 0 /* \0           */
#define Ll 1 /*   */
#define Lb 2 /*        */
#define Ln 3 /*         */
#define Lz 4 /*    */
#define Ls 5 /*         */
#define Lu 6 /* +-           */
#define Le 7 /* =            */
#define Ld 8 /* #            */
#define Lo 9 /*     */
#define Lx 10

static const u_char class[256] =
{ L0, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lb, Ll, Ll,  Lo, Ll, Lo, Lo
, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lb, Lo, Lo, Ld,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lu,  Lo, Lu, Lo, Lo
, Ln, Ln, Ln, Ln,  Ln, Ln, Ln, Ln,  Ln, Ln, Lo, Lo,  Lo, Le, Lo, Lo

, Lo, Ls, Ls, Ls,  Lz, Ls, Ls, Ls,  Lz, Ls, Ls, Ls,  Lz, Ls, Lz, Ls
, Ls, Ls, Lz, Ls,  Ls, Lz, Lz, Ls,  Ls, Ls, Ls, Lo,  Lo, Lo, Lo, Ls
, Lo, Ls, Ls, Ls,  Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls
, Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Lo,  Lo, Lo, Lo, Lo

, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lo, Lo, Lo, Ls,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lo, Lo, Lo, Ls,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo

, Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls
, Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls
, Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls
, Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls
};

static const char statenames[][3] =
{ "bg", "bf", "bs", "nl",  "nq", "tb", "tp", "sz"
, "ss", "sn", "sc", "cl",  "cq", "mc", "ml", "tq"
, "id", "in", "ii", "xx"
};

enum states
{ bg /*                                                    */
, bf /*                                                      */
, bs /*                                      */
, nl /*                                           */
, nq /*     ?                       */
, tb /*                             */
, tp /*                                           */
, sz /*   +-                                               */
, ss /*                                        */
, sn /*                                     */
, sc /*       ()          */
, cl /*                                 */
, cq /* ?                                */
, mc /*                                                 */
, ml /*                                           */
, tq /*   ?                                 */
, id /*                                     */
, in /*                                        */
, ii /*                                */
, xx
};

static const char pn[][3] = 
{ "- ", "- ", "- ", "- ", "- ", "- ", "- ", "- "
, "Cr", "Cj", "Cv", "Cm", "Ct", "Ci", "Cn", "Cb"
, "Ce", "Cc", "Cf", "Ck", "Cs", "Cd", "- ", "Cx"
};
#define Cr  0x800000  /*                                                            */
#define Cj  0x400000  /*                                                */
#define Cv  0x200000  /*                                                  */
#define Cm  0x100000  /*                                                       */
#define Ct  0x080000  /* s_num = 0   mular pgo0_style.rid                 */
#define Ci  0x040000  /*                                                                */
#define Cn  0x020000  /*                                                                    */
#define Cb  0x010000  /*                                                       */
#define Ce  0x008000  /*                                                        */
#define Cc  0x004000  /*                                                               */
#define Cp  (Cb | Cc) /*                                                */
#define Cf  0x002000  /*                                                                    */
#define Ck  0x001000  /*                                                                */
#define Cs  0x000800  /*                                                           */
#define Cd  0x000400  /*  , s_num = 0,  mular pgo0_style.str     */
#define Cx  0x000100  /*                                                                    */
#define C_state 0xFF  /*                                                           */

static u_int32_t automa[xx][Lx] =
/*       \0        \n        \b      0..9         T         a        +-      =      #     .. */
/*       L0        Ll        Lb        Ln        Lz        Ls        Lu     Le     Ld     Lo */
{{       xx,       nl,       bg,       bs,    Cf|bf,       bs,       bs,    bs,    bs,    bs }/*bg*/
,{       xx,       nl,       bs,       bs,    Cf|bf,       bs,       bs,    bs,    bs,    bs }/*bf*/
,{       xx,       nl,       bs,       bs,       bs,       bs,       bs,    bs,    bs,    bs }/*bs*/
,{       xx,       nl,       bg,       bg,       bg,       bg,       bg,    bg,    nq,    bg }/*nl*/
,{       xx,       nl,       tb,       xx,       xx,       xx,    Cd|sz,    id,    bg, Cx|xx }/*nq*/
,{    Ck|xx,    Ck|nl,       tb,    Cx|xx, Cs|Cd|tp, Cs|Cd|tp,    Cd|sz, Ck|id, Ck|xx, Ck|xx }/*tb*/
,{       xx, Cm|   cl, Cm|   sc,       tp,       tp,       tp,    Cx|xx, Cx|xx, Cx|xx, Cx|xx }/*tp*/
,{    Cx|xx,    Cx|xx,    Cx|xx,    Cx|xx,    Ci|ss,    Cx|xx,    Cx|xx, Cx|xx, Cx|xx, Cx|xx }/*sz*/
,{ Cr|Ck|xx, Cr|   cl, Cr|   sc, Cn|   sn,    Cx|xx,    Cx|xx,    Cx|xx, Cx|xx, Cx|xx, Cx|xx }/*ss*/
,{ Cr|Ck|xx, Cr|   cl, Cr|   sc, Cn|   sn,    Cx|xx,    Cx|xx,    Cx|xx, Cx|xx, Cx|xx, Cx|xx }/*sn*/
,{ Cr|Ck|xx,       cl,       sc,       sc,       sc,       sc,       sc,    sc,    sc,    sc }/*sc*/
,{    Ck|xx, Cp|Ce|ml, Cp|   mc, Cp|   mc, Cp|   mc, Cp|   mc, Cp|   mc, Cp|mc,    cq, Cp|mc }/*cl*/
,{    Ck|xx, Ck|   nl,       tb, Ck|   xx, Ck|   xx, Ck|   xx, Ck|   xx, Ck|xx, Cp|mc, Ck|xx }/*cq*/
,{ Ce|Ck|xx, Ce|Cc|ml,    Cc|mc,    Cc|mc,    Cc|mc,    Cc|mc,    Cc|mc, Cc|mc, Cc|mc, Cc|mc }/*mc*/
,{    Ck|xx, Ce|Cc|ml,    Cc|mc,    Cc|mc,    Cc|mc,    Cc|mc,    Cc|mc, Cc|mc,    tq, Cc|mc }/*ml*/
,{    Ck|xx, Ck|   nl, Ck|   tb, Ck|   xx, Ck|   xx, Ck|   xx, Ck|Cd|sz, Ck|xx, Cc|mc, Ck|xx }/*tq*/
,{       xx,       nl,       id, Ct|Cn|in, Cs|Ct|ii, Cs|Ct|ii,    Cx|xx, Cx|xx, Cx|xx, Cx|xx }/*id*/
,{    Cj|xx,    Cj|nl,    Cj|id,    Cn|in,    Cx|xx,    Cx|xx,    Cx|xx, Cx|xx, Cx|xx, Cx|xx }/*in*/
,{    Cv|xx,    Cv|nl,    Cv|id,       ii,       ii,       ii,    Cx|xx, Cx|xx, Cx|xx, Cx|xx }/*ii*/
};

static size_t msz[] = {1024, 512, 512};
static size_t asz[] = {512, 512, 512};

static void dump(pgoblin_styreg *reg_style);

static int
/**********************************************************************
 **                                                                  **/
getind(mular_descriptor *md, u_char *nm) {                          /**
 **                                                                  **
 **********************************************************************
 **    pgo0_arb   nm                   **
 **********************************************************************/
    pgo0_arb *arbp;
    u_int     i;

    for (i = 0; i < MULAR_NEXT(md); i++) {
        arbp = mular_getix(md, i);
        if  (!strcmp((char*)arbp->nm, (char*)nm)) return(i);
    }
    if  (md->flags & MULAR_FIXD) return(-1);
    i = MULAR_NEXT(md);
    arbp = mular_add(md);
    if  (!arbp) {
        /* XXXX */
    }
    arbp->nm = nm;
    return(i);
}

static int
/**********************************************************************
 **                                                                  **/
parser(pgoblin_styreg *reg_style, u_char *in) {                     /**
 **                                                                  **
 **********************************************************************/
#   define            ST     ((pgo0_style *)(reg_style->style))
    u_char           *literal = NULL; /*                */
    u_char           *l_end = NULL;   /*                                    */
    u_char           *l_beg = NULL;   /*                                     */
    mular_descriptor *ids = NULL;     /*             */
    int               ex = EX_OK;     /*                                                  */
    u_int32_t         s_num = 0;      /*                                                   */
    size_t            o_beg = 0;      /*                                         */
    u_char            s_lit = M;      /*  entity                                                 */
    int               type = 0;       /*  +(>0)  -(<0)    (&1)  entity      */
    u_int32_t         control;        /*                   */
    int               lfcount;        /*                                                 */
    size_t            offset;         /*                         */
    enum states       state;          /*                            */

#   define blin_internal_flags ((reg_style ? reg_style->flags : pgoblin_default) & BLIN_MASK)
    if  (!in) {
        ifBLIN_QX0("No pgoblin style");
        ERROUT(EX_DATAERR, EINVAL);
    }
    if  (!ST) {
        if  (!(reg_style->style = calloc(1, sizeof(pgo0_style)))) {
            ifBLIN_QX0("No memory #1");
            ERROUT(EX_OSERR, ENOMEM);
        }
        strncpy(ST->id, PGOBLIN_STYLE_ID, strlen(PGOBLIN_STYLE_ID));
    } else {
        if  (!!strncmp(ST->id, PGOBLIN_STYLE_ID, strlen(PGOBLIN_STYLE_ID))) {
            ifBLIN_QX0("Illegal style %s", ST->id);
            ERROUT(EX_DATAERR, EINVAL);
        }
        if  ((ST->flags & PGOBLIN_NTBSZMK) != PGOBLIN_NTBSIZE) {
            ifBLIN_QX0("Illegal style size %d", ST->flags & PGOBLIN_NTBSZMK);
            ERROUT(EX_DATAERR, EINVAL);
    }   }
    if  (!ST->str && !(ST->str = mular_create(MULAR_STRI | MULAR_CHAR, 3, sizeof(char*), msz))) {
        ifBLIN_QW0("mular_create #1");
        ERROUT(EX_OSERR, ENOMEM);
    }
    ST->flags = PGOBLIN_NTBSIZE;
    if  (  !ST->arb
        && !(ST->arb = mular_create( MULAR_STRI | MULAR_ZERO | MULAR_HEXU | MULAR_SWP1 | MULAR_SWP2
                                   , 3
                                   , sizeof(pgo0_arb)
                                   , asz
        )   )                      ) {
        ifBLIN_QW0("mular_create #2");
        ERROUT(EX_OSERR, ENOMEM);
    }
    ST->arb->flags &= ~MULAR_FIXD;
    blin_stateheader(BLIN_4STA24G, pn);
    lfcount = 1;
    for (state = bg, offset = 0, control = 0; state < xx; offset++, state = control & C_state) {
        if  (in[offset] == '\n') lfcount++;
        control = automa[state][class[in[offset]]];
        blin_statebody( BLIN_4STO24G
                      , pn
                      , statenames
                      , clasnames[class[in[offset]]]
                      , (char*)in
                      , offset
                      , control
                      , state
                      , 0xFFFFFFFF
                      , s_num
                      );
        if  (control & Cr) {
            u_int32_t entnum;         /*  entity                                                */
            pgo0_term term;

            if  (!(s_num < (1 << PGOBLIN_NTBSIZE))) {
                ifBLIN_QX0("Big number pgo0 line %d", lfcount);
                ERROUT(EX_DATAERR, EINVAL);
            }
            entnum = entity[s_lit - '@'];
            ifBLIN_QO1 {
                if  (type & 1) {
                    if  (ST->a[entnum].def)
                        ifBLIN_QX1( "Repeat %c%c in line %d\n"
                                  , (type > 0) ? '+' : '-', s_lit
                                  , lfcount
                                  );
                } else if ((type < 0) ? ST->a[entnum].spec[s_num].b : ST->a[entnum].spec[s_num].f) {
                    ifBLIN_QX1( "Repeat %c%c%d in line %d\n"
                              , (type > 0) ? '+' : '-', s_lit, (int)s_num
                              , lfcount
                              );
            }   }
            ifBLIN_QX5("Cr add [%"BLIN_D"]", MULAR_NEXT(ids));
            term.index = 0;
            term.arbi = 1;
            term.sign = (type < 0);
            term.type = entnum;
            term.dflt = type & 1;
            term.numb = s_num;
            *(pgo0_term*)mular_add(ids) = term;
            ifBLIN_QX5("%08X", *(u_int32_t*)mular_getix(ids, MULAR_NEXT(ids) - 1));
            l_beg = NULL;
            l_end = NULL;
        }
        if  (control & Cj) {
            if  (s_num & D_ARBI) {
                ifBLIN_QX0("Overflow line %d", lfcount);
                ERROUT(EX_DATAERR, EINVAL);
            }
            *(u_int32_t*)mular_add(ST->rid) = s_num;
        }
        if  (control & Cv) {
            u_int32_t idx;

            in[offset] = '\0';
            idx = MULAR_NEXT(ST->str);
            if  (idx & D_ARBI) {
                ifBLIN_QX0("Overflow line %d", lfcount);
                ERROUT(EX_DATAERR, EINVAL);
            }
            *(u_char**)(mular_add(ST->str)) = &in[o_beg];
            *(u_int32_t*)mular_add(ST->rid) = idx | D_ARBI;
        }
        if  (control & Cm) {
            u_int32_t arbidx;

            in[offset] = '\0';
            arbidx = getind(ST->arb, &in[o_beg]);
            if  (arbidx & D_ARBI) {
                ifBLIN_QX0("Overflow line %d", lfcount);
                ERROUT(EX_DATAERR, EINVAL);
            }
            if  (((pgo0_arb*)mular_getix(ST->arb, arbidx))->vl) {
                ifBLIN_QX1("Repeat %s in line %d", &in[o_beg], lfcount);
            }
            ifBLIN_QX5("Cm add [%"BLIN_D"]", MULAR_NEXT(ids));
            *(u_int32_t*)mular_add(ids) = arbidx;
            ifBLIN_QX5("%08X", *(u_int32_t*)mular_getix(ids, MULAR_NEXT(ids) - 1));
            l_beg = NULL;
            l_end = NULL;
        }
        if  (control & Ct) {
            s_num = 0;
            if  (!ST->rid) ST->rid = mular_create(MULAR_STRI, 3, sizeof(u_int32_t), msz);
        }
        if  (control & Ci) s_lit = in[offset];
        if  (control & Cn) {
            s_num = s_num * 10 + (in[offset] - '0');
            if  (s_num >= 0x10000000) {
                ifBLIN_QX0("Long number %u @ pgo0 line %d", s_num, lfcount);
                ERROUT(EX_DATAERR, EINVAL);
            }
            if  (type & 0xFF) type <<= 1;
        }
        if  (control & Cb && !l_beg) l_beg = literal = &in[offset];
        if  (control & Ce && l_beg) l_end = literal;
        if  (control & Cc) {
            if  (literal != &in[offset]) *literal = in[offset];
            literal++;
        }
        if  (control & Cf) {
            u_int32_t nwfl = flagvalues[entity[in[offset] - '@']];
            if  (ST->flags & nwfl & EDGE) nwfl <<= 1;
            ST->flags |= nwfl;
        }
        if  ((control & Ck) && !!ids) {
            u_int32_t  n = 0;

            if  (l_beg && l_end) {
                *l_end = '\0';
                n = MULAR_NEXT(ST->str);
                *(u_char**)(mular_add(ST->str)) = l_beg;
            }
            for (u_int k = 0; k < MULAR_NEXT(ids); k++) {
                int        entname = 0;/*  entity                                                 */
                pgo0_term *term;

                term = mular_getix(ids, k);
                ifBLIN_QO5 {
                    for (entname = 0; entname < ENTITY_LETTERS; entname++) {
                        if  (term->type == entity[entname]) break;
                    }
                    entname += '@';
                }
                if  (!(term->arbi)) {
                    ifBLIN_QX5("Ck [%u]%08X index %d", k, term->index, term->index);
                    ((pgo0_arb*)mular_getix(ST->arb, term->index))->vl = l_end ? ~n : 0;
                } else if (term->dflt) {
                    ifBLIN_QX5("Ck [%u]%08X D_DFLT D_TYPE %c", k, term->index, entname);
                    ST->a[term->type].def = l_end ? ~n : 0;
                } else {
                    ifBLIN_QX5( "Ck [%u]%08X sign=%c type=%c numb=%d"
                              , k, term->index
                              , (term->sign) ? '-' : '+'
                              , entname
                              , term->numb
                              );
                    if  (term->sign) {
                        ST->a[term->type].spec[term->numb].b = l_end ? ~n : 0;
                    } else {
                        ST->a[term->type].spec[term->numb].f = l_end ? ~n : 0;
            }   }   }
            mular_destroy(ids);
            ids = NULL;
        }
        if  (control & Cs) o_beg = offset;
        if  (control & Cd) {
            s_num = 0;
            if  (in[offset] == '-') type = -1; else type = 1;
            if  (!ids) ids = mular_create(MULAR_STRI, 3, sizeof(u_int32_t), msz);
        }
        if  (control & Cx) {
            ifBLIN_QX0("Erroneous line %d", lfcount);
            ERROUT(EX_DATAERR, EINVAL);
    }   }
out:
    if  (ST->arb) ST->arb->flags |= MULAR_FIXD;
    if  (ids) mular_destroy(ids);
    ifBLIN_QO2 dump(reg_style);
    return(ex);
#   undef blin_internal_flags
#   undef  ST
}

static void
/**********************************************************************
 **                                                                  **/
dump(pgoblin_styreg *reg_style) {                                   /**
 **                                                                  **
 **********************************************************************/
    int         ex = EX_OK;
    pgo0_style *st;
    pgo0_arb   *e;
    u_char    **c;
    u_int       u;
    int         i;
    int         j;
    int         k;
   /******************************************************/
    u_int       tmpidx;  /* Due to the error in cc -v    *
    *** Using built-in specs.                          ***
    *** Configured with: FreeBSD/amd64 system compiler ***
    *** Thread model: posix                            ***
    *** gcc version 3.4.6 [FreeBSD] 20060305           ***
    * I can't use direct                                 *
    *** mular_getix(st->str, ~st->a[i].def)            ***
    ******************************************************/

#   define blin_internal_flags ((reg_style ? reg_style->flags : pgoblin_default) & BLIN_MASK)
    if  (!reg_style || !(st = reg_style->style)) {
        printf("Dump pgo0: no style\n");
        ERROUT(EX_DATAERR, EINVAL);
    }
    if  ((st->flags & PGOBLIN_NTBSZMK) != PGOBLIN_NTBSIZE) {
        printf("Illegal pgoblin pgo0 style size %d\n", st->flags & PGOBLIN_NTBSZMK);
        ERROUT(EX_DATAERR, EINVAL);
    }
    printf("+######################################");
    for (i = 0; i < PGOBLIN_STRING_ID_LEN && st->id[i]; i++) printf("%c", st->id[i]);
    printf("\n flags=%08X %08X\n", reg_style->flags, st->flags);
    ifBLIN_QO2 {
        printf("------->str[%d]\n", (int)MULAR_NEXT(st->str));
        ifBLIN_QO3 mular_dump(st->str, stdout);
        for (u = 0; u < MULAR_NEXT(st->str); u++) {
            c = mular_getix(st->str, u);
            if  (!c) printf("[%d]\n", u);
            else if (!*c) printf("[%u]%" BLIN_X "\n", u, BLIN_I(c));
            else printf("[%u]=%s~\n", u, *c);
        }
        printf("------->arb[%d]\n", (int)MULAR_NEXT(st->arb));
        ifBLIN_QO3 mular_dump(st->arb, stdout);
        for (u = 0; u < MULAR_NEXT(st->arb); u++) {
            e = mular_getix(st->arb, u);
            if  (!e) printf("[%u]\n", u);
            else if (e->nm) printf("%u[%s]%d\n", u, e->nm, (int)~e->vl);
            else printf("%u[]%d\n", u, (int)~e->vl);
    }   }
    for (u = 0; u < MULAR_NEXT(st->arb); u++) {
        e = mular_getix(st->arb, u);
        tmpidx = ~e->vl;
        printf("# %s\n%s\n", e->nm, *(u_char**)mular_getix(st->str, tmpidx));
    }
    for (i = 0; i < M; i++) {
        for (j = 0; j < ENTITY_LETTERS; j++) if  (i == entity[j]) break;
        j += '@';
        if  (~(tmpidx = ~st->a[i].def)) {
            printf("# +%c\n%s\n", j, *(u_char**)mular_getix(st->str, tmpidx));
        }
        for (k = 0; k < (1 << PGOBLIN_NTBSIZE); k++) {
            if  (~(tmpidx = ~st->a[i].spec[k].f)) {
                printf("# +%c%d\n%s\n", j, k, *(u_char**)mular_getix(st->str, tmpidx));
            }
            if  (~(tmpidx = ~st->a[i].spec[k].b)) {
                printf("# -%c%d\n%s\n", j, k, *(u_char**)mular_getix(st->str, tmpidx));
    }   }   }
    ifBLIN_QO3 {
        for (i = 0; i < M; i++) {
            for (j = 0; j < ENTITY_LETTERS; j++) if  (i == entity[j]) printf("%c", j + '@');
            printf("%c", st->a[i].def ? '|' : '.');
            for (j = 0; j < 76; j++)
                printf( "%c"
                      , st->a[i].spec[j].f
                      ? (st->a[i].spec[j].b ? '*' : '+')
                      : (st->a[i].spec[j].b ? '-' : '.')
                      );
            printf("\n");
    }   }
    printf("-######################################\n");
    fflush(stdout);
out:
    return;
#   undef blin_internal_flags
}

static int
/**********************************************************************
 **                                                                  **/
clear(pgoblin_styreg *reg_style) {                                  /**
 **                                                                  **
 **********************************************************************/
    int         ex = EX_OK;
    pgo0_style *st;

#   define blin_internal_flags ((reg_style ? reg_style->flags : pgoblin_default) & BLIN_MASK)
    if  (reg_style) {
        if  ((st = reg_style->style)) {
            if  (!!strncmp(st->id, PGOBLIN_STYLE_ID, strlen(PGOBLIN_STYLE_ID))) {
                ifBLIN_QX0("Illegal style %s", st->id);
                ERROUT(EX_DATAERR, EINVAL);
            }
            if  ((st->flags & PGOBLIN_NTBSZMK) != PGOBLIN_NTBSIZE) {
                ifBLIN_QX0("Illegal style size %d", st->flags & PGOBLIN_NTBSZMK);
                ERROUT(EX_DATAERR, EINVAL);
            }
            if  (st->str) mular_destroy(st->str);
            if  (st->arb) mular_destroy(st->arb);
            free(st);
            st = NULL;
            reg_style->path = NULL;
            reg_style->flags &= ~PGOBLIN_STYLE_TYPE;
    }   }
out:
    return(ex);
#   undef blin_internal_flags
}

static char *
/**********************************************************************
 **                                                                  **/
get(pgoblin_styreg *reg_style, u_char *stringname) {                /**
 **                                                                  **
 **********************************************************************/
    char       *found = NULL;
    int         ex = EX_OK;
    pgo0_arb   *arbp;
    pgo0_style *st;
    int         i;
   /******************************************************/
    u_int       tmpidx;  /* Due to the error in cc -v    *
    *** Using built-in specs.                          ***
    *** Configured with: FreeBSD/amd64 system compiler ***
    *** Thread model: posix                            ***
    *** gcc version 3.4.6 [FreeBSD] 20060305           ***
    * I can't use direct                                 *
    *** mular_getix(st->str, ~arbp->vl)                ***
    ******************************************************/

#   define blin_internal_flags ((reg_style ? reg_style->flags : pgoblin_default) & BLIN_MASK)
    st = reg_style->style;
    if  (!!strncmp(st->id, PGOBLIN_STYLE_ID, strlen(PGOBLIN_STYLE_ID))) {
        ifBLIN_QX0("Illegal style %s", st->id);
        ERROUT(EX_DATAERR, EINVAL);
    }
    if  ((st->flags & PGOBLIN_NTBSZMK) != PGOBLIN_NTBSIZE) {
        ifBLIN_QX0("Illegal style size %d", st->flags & PGOBLIN_NTBSZMK);
        ERROUT(EX_DATAERR, EINVAL);
    }
    if  ((i = getind(st->arb, stringname)) >= 0) {
        arbp = mular_getix(st->arb, i);
        if  (arbp->vl) {
            tmpidx = ~(arbp->vl);
            errno = 0;
            if  (!(found = strdup(*(char**)mular_getix(st->str, tmpidx)))) errno = ENOMEM;
    }   }
out:
    return(found);
#   undef blin_internal_flags
}

static int
/**********************************************************************
 **                                                                  **/
sout(pgoblin_io *reg_out, char *c, int l) {                         /**
 **                                                                  **
 **********************************************************************/
    int ex = 0;

    if  (c) ex = reg_out->wri(reg_out->onu, c, l ? l : strlen(c));
    return(ex);
}

static int *
/**********************************************************************
 **                                                                  **/
makerenum(pgoblin_styreg *reg_style, pgoblin_mode *pq) {            /**
 **                                                                  **
 **********************************************************************/
    pgo0_style     *st = reg_style->style;
    int            *res = NULL;

#   define blin_internal_flags ((reg_style ? reg_style->flags : pgoblin_default) & BLIN_MASK)
    if  (!!st->rid) {
        res = malloc(MULAR_NEXT(st->rid) * sizeof(int));
        memset(res, -1, MULAR_NEXT(st->rid) * sizeof(int));
        for (u_int i = 0; i < MULAR_NEXT(st->rid); ++i) {
            u_int32_t *reix = mular_getix(st->rid, i);

            if  (!reix) {
                ifBLIN_QX1("Field %u is not defined", i);
            } else if (*reix & D_ARBI) {
                char **fienm = mular_getix(st->str, *reix & ~D_ARBI);

                if  (!!fienm && !!*fienm) {
                    for (int k = 0; k < pgoblin_db_resinfo(PGOBLIN_Nfields, pq); ++k) {
                        const char *reanm = pgoblin_db_subinfo(PGOBLIN_ColName, pq, k);

                        if  (!!reanm && !strcmp(*fienm, reanm)) {
                            res[i] = k;
                            goto break2;
                }   }   }
                ifBLIN_QX1("Field %u=%s~ absent", i, *fienm);
            } else {
                res[i] = *reix;
            }
break2:;
    }   }
    return(res);
#   undef blin_internal_flags
}

static char *
/*****************************************************************************************************
 **                                                                                                 **/
dsig(u_int32_t fl, mular_descriptor *s, int32_t f, int32_t b, pgo0_ent *pr) {                      /**
 **                                                                                                 **
 *****************************************************************************************************
 **       .    f       **
 **    b  ,       +f  canf   **
 ** (     )  -b  canb (       **
 ** ).    ,      .   ,  **
 **  ,      FORWARD  BAKWARD.           **
 **   ,   ,    .                 **
 **   ,   + .     ,      **
 **     .                                                              **
 *****************************************************************************************************/
    int32_t   canf = 0;
    int32_t   canb = 0;
    int32_t   can = 0;
    char    **ss = NULL;
   /******************************************************/
    u_int     tmpidx;  /* Due to the error in cc -v      *
    *** Using built-in specs.                          ***
    *** Configured with: FreeBSD/amd64 system compiler ***
    *** Thread model: posix                            ***
    *** gcc version 3.4.6 [FreeBSD] 20060305           ***
    * I can't use direct                                 *
    *** mular_getix(s, ~can)                           ***
    ******************************************************/

#   define blin_internal_flags (fl & BLIN_MASK)
    if  ((fl & FIRST) && (f < (1 << PGOBLIN_NTBSIZE))) {
        canf = pr->spec[f].f;
    }
    if  ((fl & LAST) && (b < (1 << PGOBLIN_NTBSIZE))) {
        canb = pr->spec[b].b;
    }
    if  (!canf) can = canb;
    else if (!canb) can = canf;
    else if ((fl & FORWARD) && !(fl & BAKWARD)) can = canf;
    else if (!(fl & FORWARD) && (fl & BAKWARD)) can = canb;
    else if (canb > canf) can = canb;
    else can = canf;
    if  (!can) can = pr->def;
    tmpidx = ~can;
    if  (can) ss = (char**)mular_getix(s, tmpidx);
    if  (!ss) {
        ifBLIN_QX1("tbl NULL @ f=%"BLIN_D" b=%"BLIN_D, f, b);
        return(NULL);
    } else {
        return(*ss);
    }
#   undef blin_internal_flags
}

static int
/**********************************************************************
 **                                                                  **/
tbl( pgoblin_io     *reg_out                                        /**/
   , pgoblin_styreg *reg_style                                      /**/
   , pgoblin_io     *reg_in                                         /**/
   , ssize_t         low                                            /**/
   , ssize_t         hig                                            /**/
   ) {                                                              /**
 **                                                                  **
 **********************************************************************/
    pgoblin_mode   *pq = reg_in->pq;
    ssize_t         textlength = 0; /*                                       */
    ssize_t         fullrownum = 0; /*                                     */
    int            *renum = NULL;   /*                              */
    ssize_t         hirowborder;    /* ,                 */
    char           *fillement;      /*  ,   possibleU              */
    char           *possibleU;      /*     U ,     */
    u_int32_t       flow = 0;
    ssize_t         tbwidth;        /*                                                  */
    int             ex = 0;
    u_int32_t       fwiv;
    pgo0_style     *st;
    u_int32_t       f;

#   define blin_internal_flags ((reg_style ? reg_style->flags : pgoblin_default) & BLIN_MASK)

#   define TRC(E, F, B) { /*   */                                                    \
        ifBLIN_QO3 {                                                                                  \
            int SOUT_i;                                                                               \
            fprintf(stderr, "<");                                                                     \
            for (SOUT_i = 0; SOUT_i < ENTITY_LETTERS; SOUT_i++) {                                     \
                if  ((E) == entity[SOUT_i]) fprintf(stderr, "%c", SOUT_i + '@');                      \
            }                                                                                         \
            fprintf( stderr, "%c,%d%c,%d%c>"                                                          \
                   , st->a[E].def ? '+' : '-'                                                         \
                   , (int)(F), st->a[E].spec[F].f ? '+' : '-'                                         \
                   , (int)(B), st->a[E].spec[B].b ? '+' : '-'                                         \
                   );                                                                                 \
    }   }

#   define SOUT(C, L) do { /*          */               \
        ex = sout(reg_out, (C), (L));                                                                 \
        if  (ex < 0) goto out;                                                                        \
        textlength += ex;                                                                                      \
    } while(0)

#   define VOUT(C, V) do {                                                                            \
        if  (C) {                                                                                     \
            flow |= HALWAYS;                                                                          \
            SOUT((V), 0);                                                                             \
            (V) = NULL;                                                                               \
            SOUT((C), 0);                                                                             \
        }                                                                                             \
    } while(0)

    if  (!reg_style) {
        ifBLIN_QX0("No style");
        ERROUT(EX_DATAERR, EFAULT);
    }
    st = reg_style->style;
    if  (!!strncmp(st->id, PGOBLIN_STYLE_ID, strlen(PGOBLIN_STYLE_ID))) {
        ifBLIN_QX0("Illegal style %s", st->id);
        ERROUT(EX_DATAERR, EINVAL);
    }
    if  ((st->flags & PGOBLIN_NTBSZMK) != PGOBLIN_NTBSIZE) {
        ifBLIN_QX0("Illegal style size %d", st->flags & PGOBLIN_NTBSZMK);
        ERROUT(EX_DATAERR, EINVAL);
    }
    renum = makerenum(reg_style, pq);
    if  (low >= 0) {
        flow |= FIRST | FORWARD;
        fullrownum = low;
    }
    hirowborder = fullrownum + pgoblin_db_resinfo(PGOBLIN_Ntuples, pq);
    if  ((hig == -2) || ((hig >= 0) && (hirowborder == hig))) {
        flow |= LAST | BAKWARD;
    }
    tbwidth = (!renum) ? pgoblin_db_resinfo(PGOBLIN_Nfields, pq) : MULAR_NEXT(st->rid);
    f = st->flags;
    flow &= (f >> HSHIFT) | P_A;
    fwiv = (f >> VSHIFT) | P_A;
    ifBLIN_QX2("pgo0 %08X %08X %"BLIN_D"..%"BLIN_D"x%"BLIN_D, f, flow, fullrownum, hirowborder, tbwidth);
    if  (hirowborder == 0) {
        flow &= P_A;
        if  (f & HFIRST) {
            fillement = dsig(flow | FORWARD, st->str, 0, 0, &st->a[H]);
            TRC(H, 0, 0);
            SOUT(fillement, 0);
        }
        if  (f & HLAST) {
            fillement = dsig(flow | BAKWARD, st->str, 0, 0, &st->a[H]);
            TRC(H, 0, 0);
            SOUT(fillement, 0);
        }
    } else if (tbwidth <= 0) {
        flow &= P_A;
        if  (f & HFIRST) {
            fillement = dsig(flow, st->str, 0, hirowborder, &st->a[H]);
            TRC(H, 0, hirowborder);
            SOUT(fillement, 0);
        }
        for (; fullrownum < hirowborder; fullrownum++) {
            fillement = dsig(flow, st->str, fullrownum, hirowborder - fullrownum, &st->a[H]);
            TRC(H, fullrownum, hirowborder - fullrownum);
            SOUT(fillement, 0);
            possibleU = dsig(flow, st->str, fullrownum, hirowborder - fullrownum - 1, &st->a[U]);
            TRC(U, fullrownum, hirowborder - fullrownum - 1);
            flow &= ~HALWAYS;
            if  (f & VFIRST) {
                fillement = dsig(P_A, st->str, 0, tbwidth, &st->a[V]);
                TRC(V, 0, tbwidth);
                VOUT(fillement, possibleU);
            }
            if  (f & VLAST) {
                fillement = dsig(P_A, st->str, tbwidth, 0, &st->a[V]);
                TRC(V, tbwidth, 0);
                VOUT(fillement, possibleU);
            }
            fillement = dsig(flow, st->str, fullrownum, hirowborder - fullrownum - 1, &st->a[D]);
            TRC(D, fullrownum, hirowborder - fullrownum - 1);
            if  (flow & HALWAYS) SOUT(fillement, 0);
        }
        if  (f & HLAST) {
            fillement = dsig(flow, st->str, hirowborder, 0, &st->a[H]);
            TRC(H, hirowborder, 0);
            SOUT(fillement, 0);
        }
    } else {
        for (ssize_t lorow = fullrownum; fullrownum < hirowborder; fullrownum++) {
            ssize_t backrownum = hirowborder - fullrownum;

            fillement = dsig(flow, st->str, fullrownum, backrownum, &st->a[H]);
            TRC(H, fullrownum, backrownum);
            SOUT(fillement, 0);
            possibleU = dsig(flow, st->str, fullrownum, backrownum - 1, &st->a[U]);
            TRC(U, fullrownum, backrownum - 1);
            flow = (flow & ~HALWAYS) | (st->flags & HALWAYS);
            for (int colnum = 0; colnum < tbwidth; colnum++) {
                ssize_t localrownum = fullrownum - lorow;
                ssize_t backcolnum = tbwidth - colnum;

                if  (ex < 0) goto out;
                fillement = dsig(fwiv, st->str, colnum, backcolnum, &st->a[V]);
                TRC(V, colnum, backcolnum);
                VOUT(fillement, possibleU);
                if  ( !renum
                    ? !pgoblin_db_valinfo(PGOBLIN_IsNull, pq, localrownum, colnum)
                    : !(  (renum[colnum] < 0)
                       || pgoblin_db_valinfo(PGOBLIN_IsNull, pq, localrownum, renum[colnum])
                    )  ) {
                    flow |= HALWAYS;
                    SOUT(possibleU, 0);
                    possibleU = NULL;
                    fillement = dsig(fwiv, st->str, colnum, backcolnum - 1, &st->a[L]);
                    TRC(L, colnum, backcolnum - 1);
                    SOUT(fillement, 0);
                    if  (!renum) {
                        SOUT( pgoblin_db_getvalue(pq, localrownum, colnum)
                            , pgoblin_db_valinfo(PGOBLIN_Length, pq, localrownum, colnum)
                            );
                    } else {
                        SOUT( pgoblin_db_getvalue(pq, localrownum, renum[colnum])
                            , pgoblin_db_valinfo(PGOBLIN_Length, pq, localrownum, renum[colnum])
                            );
                    }
                    fillement = dsig(fwiv, st->str, colnum, backcolnum - 1, &st->a[R]);
                    TRC(R, colnum, backcolnum - 1);
                    SOUT(fillement, 0);
                } else if (!!(fillement = dsig(fwiv, st->str, colnum, backcolnum - 1, &st->a[N]))) {
                    TRC(N, colnum, backcolnum);
                    VOUT(fillement, possibleU);
                } else if (f & VALWAYS) {
                    fillement = dsig(fwiv, st->str, colnum, backcolnum - 1, &st->a[L]);
                    TRC(L, colnum, backcolnum);
                    VOUT(fillement, possibleU);
                    fillement = dsig(fwiv, st->str, colnum, backcolnum - 1, &st->a[R]);
                    TRC(R, colnum, backcolnum - 1);
                    VOUT(fillement, possibleU);
            }   }
            fillement = dsig(fwiv, st->str, tbwidth, 0, &st->a[V]);
            TRC(V, tbwidth, 0);
            VOUT(fillement, possibleU);
            fillement = dsig(flow, st->str, fullrownum, backrownum - 1, &st->a[D]);
            TRC(D, fullrownum, backrownum - 1);
            if  (flow & HALWAYS) SOUT(fillement, 0);
        }
        if  (flow & LAST) {
            fillement = dsig(flow, st->str, hirowborder, 0, &st->a[H]);
            TRC(H, hirowborder, 0);
            SOUT(fillement, 0);
    }   }
out:
    free(renum);
    return((ex < 0) ? ex : textlength);
#   undef blin_internal_flags
}

static int
/**********************************************************************
 **                                                                  **/
init(u_int32_t flags) {                                             /**
 **                                                                  **
 **********************************************************************/
    return(0);
}

static int
/**********************************************************************
 **                                                                  **/
fini() {                                                            /**
 **                                                                  **
 **********************************************************************/
    return(0);
}

pgoblin_styles PGOBLIN_MODULE =
{ 0
, "5pGoblin-" VERS
, "pgo0\0" VERS
, 0
, init
, fini
, parser
, dump
, clear
, get
, tbl
, NULL
, NULL
, NULL
, NULL
};
